package GTM;

# ------------------------------------------------------------------------

#Copyright (C) 2005 Enrique Amig and Jess Gimnez

#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.

#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.

#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# ------------------------------------------------------------------------

use strict;
use Data::Dumper;
use IQ::IQXML;
use IQ::COMMON;
use File::ReadBackwards;

use vars qw($GTMEXT $rGTM $TGTM);

$GTM::GTMEXT = "GTM";
$GTM::TGTM = "GTM/gtm-1.2-binary/gtm.jar +s +d";   # version 1.2
$GTM::rGTM = { "$GTM::GTMEXT-1" => 1, "$GTM::GTMEXT-2" => 1, "$GTM::GTMEXT-3" => 1 };

sub read_gtm {
   #description _ read GTM value from report file
   #param1  _ report filename
   #@return _ GTM score

   my $report = shift;

   my $REPORT = File::ReadBackwards->new($report) or die "Couldn't open input file: $report\n";
   my $gtm = $REPORT->readline;
   $REPORT->close();
   chomp($gtm);
   $gtm =~ s/^ +//;

   return $gtm;
}

sub read_gtm_segments {
   #description _ read GTM value from report file (for all segments)
   #param1  _ report filename
   #@return _ gtm F1 score list

   my $report = shift;

   my @lgtm;

   open(AUX, "< $report") or die "couldn't open file: $report\n";
   while (my $aux = <AUX>) {
      if ($aux =~ /^ \".*/) {
         chomp($aux);
         my @laux = split(/ +/, $aux);
         push(@lgtm, $laux[2]);
      }
   }
   close(AUX);

   return \@lgtm;
}

sub computeMultiGTM {
   #description _ computes GTM F1 (by calling Proteus java gtm) -> e = 1..3 (multiple references)
   #param1 _ source file
   #param2 _ candidate file
   #param3 _ reference file(s) [hash reference]
   #param4 _ remake reports? (1 - yes :: 0 - no)
   #param5 _ TARGET NAME (base filename)
   #param6 _ REF NAME (base filename)
   #param7 _ tools
   #param8 _ e parameter
   #param9 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $Href = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $e = shift;
   my $verbose = shift;

   my $toolGTM = "java -jar $tools/$GTM::TGTM";
   my @LrefGTMsgml;
   foreach my $r (keys %{$Href}) {
       my $ref = $Href->{$r};
       my $refGTMsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::REFEXT.$GTMEXT.$COMMON::SGMLEXT";
       if ((!(-e $refGTMsgml)) or $remakeREPORTS) { IQXML::GTM_f_create_mteval_doc($ref, $refGTMsgml, $verbose); }
       push(@LrefGTMsgml, $refGTMsgml);
   }
   my $outGTMsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::SYSEXT.$GTMEXT.$COMMON::SGMLEXT";
   if ((!(-e $outGTMsgml)) or $remakeREPORTS) { IQXML::GTM_f_create_mteval_doc($out, $outGTMsgml, $verbose); }
   my $reportGTM = "./$COMMON::TMP/".rand($COMMON::NRAND).".$GTMEXT-$e.$COMMON::REPORTEXT";
   if ($verbose > 1) { print STDERR "building $reportGTM...\n"; }
   system("$toolGTM -e $e $outGTMsgml ".join(" ", @LrefGTMsgml)." > $reportGTM");
   my $SYS = read_gtm($reportGTM);
   my $SEG = read_gtm_segments($reportGTM);
   system("rm -f $reportGTM");
   foreach my $r (@LrefGTMsgml) { if (-e $r) { system("rm -f $r"); } }
   if (-e $outGTMsgml) { system "rm -f $outGTMsgml"; }

   return($SYS, $SEG);
}

sub doMultiGTM {
   #description _ computes GTM F1 (by calling Proteus java gtm) -> e = 1..3 (multiple references)
   #param1  _ source file
   #param2  _ candidate file
   #param3  _ reference file(s) [hash reference]
   #param4  _ remake reports? (1 - yes :: 0 - no)
   #param5  _ TARGET NAME (base filename)
   #param6  _ REF NAME (base filename)
   #param7  _ tools
   #param8  _ metrics to compute
   #param9  _ optional prefix
   #param10 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $Href = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $M = shift;
   my $prefix = shift;
   my $verbose = shift;

   my $GO = 0; my $i = 0;
   my @mGTM = keys %{$GTM::rGTM};
   while (($i < scalar(@mGTM)) and (!$GO)) { if (exists($M->{$prefix.$mGTM[$i]})) { $GO = 1; } $i++; }

   if ($GO) {
      if ($verbose == 1) { print STDERR "GTM.."; }
      my $reportGTM1xml = "./$TGTname/$REFname/$prefix$GTMEXT-1.$COMMON::XMLEXT";
      my $reportGTM2xml = "./$TGTname/$REFname/$prefix$GTMEXT-2.$COMMON::XMLEXT";
      my $reportGTM3xml = "./$TGTname/$REFname/$prefix$GTMEXT-3.$COMMON::XMLEXT";
      if ((!(-e $reportGTM1xml) and !(-e $reportGTM1xml.".$COMMON::GZEXT") and ($M->{"$GTMEXT-1"})) or $remakeREPORTS) {
         my ($SYS, $SEG) = GTM::computeMultiGTM($src, $out, $Href, $remakeREPORTS, $TGTname, $REFname, $tools, 1, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$GTMEXT-1", $SYS, $SEG, $reportGTM1xml, $verbose);
      }
      if ((!(-e $reportGTM2xml) and !(-e $reportGTM2xml.".$COMMON::GZEXT") and ($M->{"$GTMEXT-2"})) or $remakeREPORTS) {
         my ($SYS, $SEG) = GTM::computeMultiGTM($src, $out, $Href, $remakeREPORTS, $TGTname, $REFname, $tools, 2, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$GTMEXT-2", $SYS, $SEG, $reportGTM2xml, $verbose);
      }
      if ((!(-e $reportGTM3xml) and !(-e $reportGTM3xml.".$COMMON::GZEXT") and ($M->{"$GTMEXT-3"})) or $remakeREPORTS) {
         my ($SYS, $SEG) = GTM::computeMultiGTM($src, $out, $Href, $remakeREPORTS, $TGTname, $REFname, $tools, 3, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$GTMEXT-3", $SYS, $SEG, $reportGTM3xml, $verbose);
      }
   }
}

sub computeGTM {
   #description _ computes GTM F1 (by calling Proteus java gtm) -> e = 1..3 (single reference)
   #param1 _ source file
   #param2 _ candidate file
   #param3 _ reference file
   #param4 _ remake reports? (1 - yes :: 0 - no)
   #param5 _ TARGET NAME (base filename)
   #param6 _ REF NAME (base filename)
   #param7 _ tools
   #param8 _ e parameter
   #param9 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $ref = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $e = shift;
   my $verbose = shift;

   my $toolGTM = "java -jar $tools/$GTM::TGTM";
   my $refGTMsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::REFEXT.$GTMEXT.$COMMON::SGMLEXT";
   my $outGTMsgml = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::SYSEXT.$GTMEXT.$COMMON::SGMLEXT";
   if ((!(-e $refGTMsgml)) or $remakeREPORTS) { IQXML::GTM_f_create_mteval_doc($ref, $refGTMsgml, $verbose); }
   if ((!(-e $outGTMsgml)) or $remakeREPORTS) { IQXML::GTM_f_create_mteval_doc($out, $outGTMsgml, $verbose); }
   my $reportGTM = "./$COMMON::TMP/".rand($COMMON::NRAND).".$COMMON::SYSEXT.$GTMEXT-$e.$COMMON::REPORTEXT";
   if ($verbose > 1) { print STDERR "building $reportGTM...\n"; }
   system("$toolGTM -e $e $outGTMsgml $refGTMsgml > $reportGTM");
   my $SYS = read_gtm($reportGTM);
   my $SEG = read_gtm_segments($reportGTM);
   system("rm -f $reportGTM");

   if (-e $refGTMsgml) { system "rm -f $refGTMsgml"; }
   if (-e $outGTMsgml) { system "rm -f $outGTMsgml"; }

   return($SYS, $SEG);
}

sub doGTM {
   #description _ computes GTM F1 (by calling Proteus java gtm) -> e = 1..3 (single reference)
   #param1  _ source file
   #param2  _ candidate file
   #param3  _ reference file(s) [hash reference]
   #param4  _ remake reports? (1 - yes :: 0 - no)
   #param5  _ TARGET NAME (base filename)
   #param6  _ REF NAME (base filename)
   #param7  _ tools
   #param8  _ metrics to compute
   #param9  _ optional prefix
   #param10 _ verbosity (0/1)

   my $src = shift;
   my $out = shift;
   my $ref = shift;
   my $remakeREPORTS = shift;
   my $TGTname = shift;
   my $REFname = shift;
   my $tools = shift;
   my $M = shift;
   my $prefix = shift;
   my $verbose = shift;

   my $GO = 0; my $i = 0;
   my @mGTM = keys %{$GTM::rGTM};
   while (($i < scalar(@mGTM)) and (!$GO)) { if (exists($M->{$prefix.$mGTM[$i]})) { $GO = 1; } $i++; }

   if ($GO) {
      if ($verbose == 1) { print STDERR "GTM.."; }
      my $reportGTM1xml = "./$TGTname/$REFname/$prefix$GTMEXT-1.$COMMON::XMLEXT";
      my $reportGTM2xml = "./$TGTname/$REFname/$prefix$GTMEXT-2.$COMMON::XMLEXT";
      my $reportGTM3xml = "./$TGTname/$REFname/$prefix$GTMEXT-3.$COMMON::XMLEXT";
      if ((!(-e $reportGTM1xml) and !(-e $reportGTM1xml.".$COMMON::GZEXT") and ($M->{"$GTMEXT-1"})) or $remakeREPORTS) {
         my ($SYS, $SEG) = GTM::computeGTM($src, $out, $ref, $remakeREPORTS, $TGTname, $REFname, $tools, 1, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$GTMEXT-1", $SYS, $SEG, $reportGTM1xml, $verbose);
      }
      if ((!(-e $reportGTM2xml) and !(-e $reportGTM2xml.".$COMMON::GZEXT") and ($M->{"$GTMEXT-2"})) or $remakeREPORTS) {
         my ($SYS, $SEG) = GTM::computeGTM($src, $out, $ref, $remakeREPORTS, $TGTname, $REFname, $tools, 2, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$GTMEXT-2", $SYS, $SEG, $reportGTM2xml, $verbose);
      }
      if ((!(-e $reportGTM3xml) and !(-e $reportGTM3xml.".$COMMON::GZEXT") and ($M->{"$GTMEXT-3"})) or $remakeREPORTS) {
         my ($SYS, $SEG) = GTM::computeGTM($src, $out, $ref, $remakeREPORTS, $TGTname, $REFname, $tools, 3, $verbose);
         IQXML::write_report($TGTname, $REFname, "$prefix$GTMEXT-3", $SYS, $SEG, $reportGTM3xml, $verbose);
      }
   }
}

1;
