package QARLA;

# ------------------------------------------------------------------------

#Copyright (C) 2005 Enrique Amig and Jess Gimnez

#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.

#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.

#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# ------------------------------------------------------------------------

use strict;
use Data::Dumper;
use IQ::COMMON;
use IQ::IQXML;

sub start {
   #description _ given a set of metrics, systems and human references reads similarities
   #              for a given set of segments.
   #param1  _ set of metrics
   #param2  _ set of segments (topics)
   #param3  _ set of systems
   #param4  _ set of references (models)
   #param5  _ read additional similarities for JACK computation
   #param6  _ verbosity (0/1)
   #@return _ IQ structure

   my $M = shift;
   my $T = shift;
   my $S = shift;
   my $R = shift;
   my $doJACK = shift;
   my $verbose = shift;

   if ($verbose) { print STDERR "[INSIDE QARLA] READING SIMILARITIES... "; }
   my %IQ;

   my @topics;
   foreach my $t (@{$T}) { $topics[$t] = 1; }

   if ($verbose == 1) { print STDERR "["; }

   foreach my $m (@{$M}) {
      if ($verbose == 1) { print STDERR "$m.."; }
      foreach my $r1 (@{$R}) {
         foreach my $r2 (@{$R}) {
	    if ($r1 ne $r2) {
               IQXML::read_report($r1, $r2, $m, "./$r1/$r2/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
	    }
         }
      }
      foreach my $s (@{$S}) {
         foreach my $r (@{$R}) {
            IQXML::read_report($s, $r, $m, "./$s/$r/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
            #if ($doJACK) {
            #   IQXML::read_report($r, $s, $m, "./$r/$s/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
	    #}
         }
      }

      if ($doJACK) {
         foreach my $s1 (@{$S}) {
            foreach my $s2 (@{$S}) {
               if ($s1 ne $s2) {
                  IQXML::read_report($s1, $s2, $m, "./$s1/$s2/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
               }
            }
         }
      }
   }

   if ($verbose == 1) { print STDERR "]\n"; }

   my $NSEGMENTS;
   if (scalar($M > 0)) { $NSEGMENTS = scalar(@{$IQ{$M->[0]}}); }
   else { $NSEGMENTS = 0; }

   return (\%IQ, $NSEGMENTS);
}


sub load {
   #description _ given a set of metrics, systems and human references loads similarities
   #              for a given set of segments.
   #param1  _ set of metrics
   #param2  _ set of segments (topics)
   #param3  _ set of systems
   #param4  _ set of references (models)
   #param5  _ read additional similarities for JACK computation
   #param6  _ verbosity (0/1)
   #@return _ IQ structure

   my $M = shift;
   my $T = shift;
   my $S = shift;
   my $R = shift;
   my $doJACK = shift;
   my $verbose = shift;

   #if ($verbose) { print STDERR "[INSIDE QARLA] READING SIMILARITIES... "; }
   my %IQ;

   my @topics;
   foreach my $t (@{$T}) { $topics[$t] = 1; }

   if ($verbose == 1) { print STDERR "["; }

   foreach my $m (@{$M}) {
      if ($verbose == 1) { print STDERR "$m.."; }
      foreach my $r1 (@{$R}) {
         foreach my $r2 (@{$R}) {
	    if ($r1 ne $r2) {
               IQXML::read_report($r1, $r2, $m, "./$r1/$r2/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
	    }
         }
      }
      foreach my $s (@{$S}) {
         foreach my $r (@{$R}) {
            IQXML::read_report($s, $r, $m, "./$s/$r/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
            #if ($doJACK) {
            #   IQXML::read_report($r, $s, $m, "./$r/$s/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
	    #}
         }
      }

      if ($doJACK) {
         foreach my $s1 (@{$S}) {
            foreach my $s2 (@{$S}) {
               if ($s1 ne $s2) {
                  IQXML::read_report($s1, $s2, $m, "./$s1/$s2/$m.$COMMON::XMLEXT", \%IQ, \@topics, 'seg', $verbose);
               }
            }
         }
      }
   }

   if ($verbose == 1) { print STDERR "]\n"; }

   my $NSEGMENTS;
   if (scalar($M > 0)) { $NSEGMENTS = scalar(@{$IQ{$M->[0]}}); }
   else { $NSEGMENTS = 0; }

   return \%IQ;
}

sub reload {
   #description _ given a set of metrics, systems and human references reloads (only if necessary) similarities
   #              for a given set of segments.
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ read additional similarities for JACK computation
   #param7  _ verbosity (0/1)

   my $IQ = shift;
   my $M = shift;
   my $T = shift;
   my $S = shift;
   my $R = shift;
   my $doJACK = shift;
   my $verbose = shift;

   #if ($verbose) { print STDERR "[INSIDE QARLA] READING SIMILARITIES... "; }

   my @topics;
   foreach my $t (@{$T}) { $topics[$t] = 1; }

   if ($verbose > 1) { print STDERR "["; }

   foreach my $m (@{$M}) {
      if ($verbose > 1) { print STDERR "$m.."; }
      foreach my $r1 (@{$R}) {
         foreach my $r2 (@{$R}) {
	    if ($r1 ne $r2) {
	       if (!(exists($IQ->{$m}->[1]->{$r1}->{$r2}))) {
                  IQXML::read_report($r1, $r2, $m, "./$r1/$r2/$m.$COMMON::XMLEXT", $IQ, \@topics, 'seg', $verbose);
	       }
	    }
         }
      }
      foreach my $s (@{$S}) {
         foreach my $r (@{$R}) {
            if (!(exists($IQ->{$m}->[1]->{$s}->{$r}))) {
               IQXML::read_report($s, $r, $m, "./$s/$r/$m.$COMMON::XMLEXT", $IQ, \@topics, 'seg', $verbose);
	    }
            #if ($doJACK) {
	       #if (!(exists($IQ->{$m}->[1]->{$s}->{$r}))) {

               #   IQXML::read_report($r, $s, $m, "./$r/$s/$m.$COMMON::XMLEXT", $IQ, \@topics, 'seg', $verbose);
	       #}
	    #}
         }
      }

      if ($doJACK) {
         foreach my $s1 (@{$S}) {
            foreach my $s2 (@{$S}) {
               if ($s1 ne $s2) {
  	          if (!(exists($IQ->{$m}->[1]->{$s1}->{$s2}))) {
                     IQXML::read_report($s1, $s2, $m, "./$s1/$s2/$m.$COMMON::XMLEXT", $IQ, \@topics, 'seg', $verbose);
		  }
               }
            }
         }
      }
   }

   if ($verbose > 1) { print STDERR "]\n"; }

   my $NSEGMENTS;
   if (scalar($M > 0)) { $NSEGMENTS = scalar(@{$IQ->{$M->[0]}}); }
   else { $NSEGMENTS = 0; }
}

sub unload {
   #description _ given a set of metrics unloads similarities.
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ verbosity (0/1)

   my $IQ = shift;
   my $M = shift;
   my $verbose = shift;

   foreach my $m (@{$M}) {
       $IQ->{$m} = ();
   }
}

	
##############################################
#####   KING
##############################################

sub KING {
   #description _ compute KING
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $K=0;
   my $samples=0;

   foreach my $model (@{$models}) {
      #%{$IQ} = ();  #empty IQ structure
      my @modelsCV;
      foreach my $m (@{$models}) { if ($m ne $model) { push(@modelsCV, $m); } }         
      my %Qpeer; my %QpeerTOPICS;
      foreach my $peer (@{$peers}) { 
         my @lpeer = ($peer);
         my ($Q, $QTOPICS) = QUEEN($IQ, $metrics, $topics, \@lpeer, \@modelsCV, $verbose);
         $Qpeer{$peer} = $Q; $QpeerTOPICS{$peer} = $QTOPICS;
      }
      my @lmodel = ($model);
      my ($Qmodel, $QmodelTOPICS) = QUEEN($IQ, $metrics, $topics, \@lmodel, \@modelsCV, $verbose);

      foreach my $topic (@{$topics}) {
         $samples++;
         my $conditionK = 1;
         foreach my $peer (@{$peers}) { 
	    if ($QpeerTOPICS{$peer}->[$topic] >= $QmodelTOPICS->[$topic]) { $conditionK = 0; }
         }
         if ($conditionK == 1) { $K++; }
      }
   }

   if ($samples == 0) { return 0; }
   else { return $K / $samples; }
}

sub KING_TT {
   #description _ compute KING
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $K=0;
   my $samples=0;

   foreach my $model (@{$models}) {
      #%{$IQ} = ();  #empty IQ structure
      my @modelsCV;
      foreach my $m (@{$models}) { if ($m ne $model) { push(@modelsCV, $m); } }         
      my %Qpeer; my %QpeerTOPICS;
      foreach my $peer (@{$peers}) { 
         my @lpeer = ($peer);
         my ($Q, $QTOPICS) = QUEEN_TT($IQ, $metrics, $topics, $topics, \@lpeer, \@modelsCV, $verbose);
         $Qpeer{$peer} = $Q; $QpeerTOPICS{$peer} = $QTOPICS;
      }
      my @lmodel = ($model);
      my ($Qmodel, $QmodelTOPICS) = QUEEN_TT($IQ, $metrics, $topics, $topics, \@lmodel, \@modelsCV, $verbose);

      foreach my $topic (@{$topics}) {
         $samples++;  #(topics * models)
         my $conditionK = 1;
         foreach my $peer (@{$peers}) { 
	    if ($QpeerTOPICS{$peer}->[$topic] >= $QmodelTOPICS->[$topic]) { $conditionK = 0; }
         }
         if ($conditionK == 1) { $K++; }
      }
   }

   if ($samples == 0) { return 0; }
   else { return $K / $samples; }
}

########################################
##### QUEEN
#######################################

sub QUEEN {
#sub QUEEN_new {
   #description _ compute QUEEN
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)
   #@return _ (QUEEN, QUEEN-per-topic)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   #print Dumper($metrics);
   #print Dumper($topics);
   #print Dumper($peers);
   #print Dumper($models);

   my @QTOPICS;
   my @STOPICS;

   #my $IQ = QARLA::load($metrics, $topics, $peers, $models, 0, $verbose);  # high MEMORY USAGE

   foreach my $peer (@{$peers}) {
      my @lpeers = ($peer);
      # fine MEMORY USAGE
      QARLA::reload($IQ, $metrics, $topics, \@lpeers, $models, 0, $verbose);
      #$IQ = QARLA::load($metrics, $topics, \@lpeers, $models, 0, $verbose);
      foreach my $model0 (@{$models}) {	 
         if ($peer ne $model0) { 	   
            foreach my $model1 (@{$models}) { 
               if ($peer ne $model1) { 	   
	          foreach my $model2 (@{$models}){
 	             if (($peer ne $model2) && ($model1 ne $model2)) {
                        my @lmodels = ($model0, $model1, $model2);
                        # low MEMORY USAGE
                        #my $IQ = QARLA::load($metrics,$topics, \@lpeers, \@lmodels, 0,$verbose);
                        foreach my $topic (@{$topics}) {
  	                   $STOPICS[$topic]++;
                           my $existsMetric=0;
	   	           foreach my $metric (@{$metrics}) {
			      if ($IQ->{$metric}->[$topic]->{$peer}->{$model0} <
			          $IQ->{$metric}->[$topic]->{$model1}->{$model2}) { $existsMetric = 1; }

			   }
		 	   if ($existsMetric == 0) { $QTOPICS[$topic]++; }
		        }
		     }
		  }
	       }
	    }
	 }
      }
   }

   my $Q=0;
   foreach my $topic (@{$topics}) {
      if ($STOPICS[$topic] == 0) { $QTOPICS[$topic] = 0; }
      else { $QTOPICS[$topic] = $QTOPICS[$topic] / $STOPICS[$topic]; }
      $Q += $QTOPICS[$topic];
   }

   if (scalar(@{$topics}) == 0) { $Q = 0; }
   else { $Q /= scalar(@{$topics}); }

   return ($Q, \@QTOPICS);
}

sub QUEEN_TT {
   #description _ compute QUEEN tras-topic
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of all segments (all topics)
   #param5  _ set of systems
   #param6  _ set of references (models)
   #param7  _ verbosity (0/1)
   #@return _ (QUEEN, QUEEN-per-topic)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $alltopics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   #print Dumper($metrics);
   #print Dumper($topics);
   #print Dumper($peers);
   #print Dumper($models);

   my @QTOPICS;
   my @STOPICS;

   #my $IQ = QARLA::load($metrics, $topics, $peers, $models, 0, $verbose);  # high MEMORY USAGE

   foreach my $peer (@{$peers}) {
      my @lpeers = ($peer);
      # fine MEMORY USAGE
      QARLA::reload($IQ, $metrics, $topics, \@lpeers, $models, 0, $verbose);
      #$IQ = QARLA::load($metrics, $topics, \@lpeers, $models, 0, $verbose);
      foreach my $model0 (@{$models}) {	 
         if ($peer ne $model0) { 	   
            foreach my $topic2 (@{$alltopics}) {  # <---- TRASTOPIC!!
               foreach my $model1 (@{$models}) { 
                  if ($peer ne $model1) { 	   
	             foreach my $model2 (@{$models}){
 	                if (($peer ne $model2) && ($model1 ne $model2)) {
                           my @lmodels = ($model0, $model1, $model2);
                           # low MEMORY USAGE
                           #my $IQ = QARLA::load($metrics,$topics, \@lpeers, \@lmodels, 0,$verbose);
                           foreach my $topic (@{$topics}) {
  	                      $STOPICS[$topic]++;
                              my $existsMetric=0;
	   	  	      foreach my $metric (@{$metrics}) {
			         if ($IQ->{$metric}->[$topic]->{$peer}->{$model0} <
			             $IQ->{$metric}->[$topic2]->{$model1}->{$model2}) { $existsMetric = 1; }
  	      	              }	       
		 	      if ($existsMetric == 0) { $QTOPICS[$topic]++; }
			   }
		        }
		     }
		  }
	       }
	    } # <----
	 }
      }
   }

   my $Q=0;
   foreach my $topic (@{$topics}) {
      if ($STOPICS[$topic] == 0) { $QTOPICS[$topic] = 0; }
      else { $QTOPICS[$topic] = $QTOPICS[$topic] / $STOPICS[$topic]; }
      $Q += $QTOPICS[$topic];
   }

   if (scalar(@{$topics}) == 0) { $Q = 0; }
   else { $Q /= scalar(@{$topics}); }

   return ($Q, \@QTOPICS);
}

##########################################
###  JACK
###########################################

sub JACK {
#sub JACK_new {
   #description _ compute JACK
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $J=0;
   my $samples=0;

   my %Qpeer; my %QpeerTOPICS;
   foreach my $peer (@{$peers}) { 
      my @lpeer = ($peer);
      my ($Q, $QTOPICS) = QUEEN($IQ, $metrics, $topics, \@lpeer, $models, $verbose);
      $Qpeer{$peer} = $Q; $QpeerTOPICS{$peer} = $QTOPICS;
   }

   #%{$IQ} = ();  #empty IQ structure
   #print Dumper $IQ;

   foreach my $model (@{$models}) {
      my @lmodel = ($model);
      QARLA::reload($IQ, $metrics, $topics, $peers, \@lmodel, 1, $verbose);      # fine MEMORY USAGE
      foreach my $topic (@{$topics}) {
	 $samples++;
	 my $found=0;
         my $i = 0;
         while (($i <  scalar(@{$peers})) and (!$found)) {
	    if ($QpeerTOPICS{$peers->[$i]}->[$topic] > 0) {
               my $j = 0;
               while (($j <  scalar(@{$peers})) and (!$found)) {
      	          if (($i != $j) and ($QpeerTOPICS{$peers->[$j]}->[$topic] > 0)) {
		     my $conditionJ = 1;
   	      	     foreach my $metric (@{$metrics}) {
			print "MODEL = $model :: TOPIC = $topic -------------------------\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}, "\n";
                        if (($IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}) or
                           ($IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}))
                           { $conditionJ = 0; }
		     }
  		     if ($conditionJ == 1){ $found = 1; }
		  }
		  $j++;
	       }
	    }
	    $i++;
	 }
	 if ($found == 1){ $J++; }
      }
   }

#   print "--> J = $J / $samples!!\n";

   if ($samples == 0) { return 0; }
   else { return $J / $samples; }
}

#sub JACK {
sub JACK_old {
   #description _ compute JACK
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $J=0;
   my $samples=0;

   my $IQ = QARLA::load($metrics, $topics, $peers, $models, 1, $verbose);  # high MEMORY USAGE

   foreach my $topic (@{$topics}) {
      my @ltopic = ($topic);
      my %Qpeer;
      foreach my $peer (@{$peers}) {
	 my @lpeer = ($peer);
	 my ($Q, $lQ) = QUEEN($IQ, $metrics, \@ltopic, \@lpeer, $models, $verbose);
         $Qpeer{$peer} = $Q;
      }

      foreach my $model (@{$models}) {
	 $samples++;			  
	 my $found=0;
         my $i = 0;
         while (($i <  scalar(@{$peers})) and (!$found)) {
	    if ($Qpeer{$peers->[$i]} > 0) {
               my $j = 0;
               while (($j <  scalar(@{$peers})) and (!$found)) {
      	          if (($i != $j) and ($Qpeer{$peers->[$j]} > 0)) {
		     my $conditionJ = 1;
   	      	     foreach my $metric (@{$metrics}) {
			print "MODEL = $model :: TOPIC = $topic -------------------------\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}, "\n";
                        if (($IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}) or
                           ($IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}))
                           { $conditionJ = 0; }
		     }
  		     if ($conditionJ == 1){ $found = 1; }
		  }
		  $j++;
	       }
	    }
	    $i++;
	 }
	 if ($found == 1){ $J++; }
      }
   }

#   print "--> J = $J / $samples!!\n";

   if ($samples == 0) { return 0; }
   else { return $J / $samples; }
}

sub JACK_TT {
#sub JACK_TT_new {
   #description _ compute JACK
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $J=0;
   my $samples=0;

   my %Qpeer; my %QpeerTOPICS;
   foreach my $peer (@{$peers}) { 
      my @lpeer = ($peer);
      my ($Q, $QTOPICS) = QUEEN_TT($IQ, $metrics, $topics, $topics, \@lpeer, $models, $verbose);
      $Qpeer{$peer} = $Q; $QpeerTOPICS{$peer} = $QTOPICS;
   }

   #%{$IQ} = ();  #empty IQ structure
   #print Dumper $IQ;

   foreach my $model (@{$models}) {
      my @lmodel = ($model);
      QARLA::reload($IQ, $metrics, $topics, $peers, \@lmodel, 1, $verbose);      # fine MEMORY USAGE
      foreach my $topic (@{$topics}) {
	 $samples++;
	 my $found=0;
         my $i = 0;
         while (($i <  scalar(@{$peers})) and (!$found)) {
	    if ($QpeerTOPICS{$peers->[$i]}->[$topic] > 0) {
               my $j = 0;
               while (($j <  scalar(@{$peers})) and (!$found)) {
      	          if (($i != $j) and ($QpeerTOPICS{$peers->[$j]}->[$topic] > 0)) {
		     my $conditionJ = 1;
   	      	     foreach my $metric (@{$metrics}) {
			print "MODEL = $model :: TOPIC = $topic -------------------------\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}, "\n";
                        if (($IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}) or
                           ($IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}))
                           { $conditionJ = 0; }
		     }
  		     if ($conditionJ == 1){ $found = 1; }
		  }
		  $j++;
	       }
	    }
	    $i++;
	 }
	 if ($found == 1){ $J++; }
      }
   }

#   print "--> J = $J / $samples!!\n";

   if ($samples == 0) { return 0; }
   else { return $J / $samples; }
}

#sub JACK_TT {
sub JACK_TT_old {
   #description _ compute JACK tras-topic
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $J=0;
   my $samples=0;

   my $IQ = QARLA::load($metrics, $topics, $peers, $models, 1, $verbose);  # high MEMORY USAGE

   foreach my $topic (@{$topics}) {
      my @ltopic = ($topic);
      my %Qpeer;
      foreach my $peer (@{$peers}) {
	 my @lpeer = ($peer);
         my ($Q, $lQ) = QUEEN_TT($IQ, $metrics, \@ltopic, $topics, \@lpeer, $models, $verbose);
         $Qpeer{$peer} = $Q;
      }

#      print Dumper(\%Qpeer);

      foreach my $model (@{$models}) {
	 $samples++;			  
	 my $found=0;
         my $i = 0;
         while (($i <  scalar(@{$peers})) and (!$found)) {
	    if ($Qpeer{$peers->[$i]} > 0) {
               my $j = 0;
               while (($j <  scalar(@{$peers})) and (!$found)) {
      	          if (($i != $j) and ($Qpeer{$peers->[$j]} > 0)) {
		     my $conditionJ = 1;
   	      	     foreach my $metric (@{$metrics}) {
			print "MODEL = $model :: TOPIC = $topic -------------------------\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model}, "\n";
                        print $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}, "\n";
                        if (($IQ->{$metric}->[$topic]->{$peers->[$i]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$i]}->{$peers->[$j]}) or
                           ($IQ->{$metric}->[$topic]->{$peers->[$j]}->{$model} <
                             $IQ->{$metric}->[$topic]->{$peers->[$j]}->{$peers->[$i]}))
                           { $conditionJ = 0; }
		     }
  		     if ($conditionJ == 1){ $found = 1; }
		  }
		  $j++;
	       }
	    }
	    $i++;
	 }
	 if ($found == 1){ $J++; }
      }
   }

   print "--> J = $J / $samples\n!!";

   if ($samples == 0) { return 0; }
   else { return $J / $samples; }
}


###########################################
###  TUNING (KING optimization) -- simple
###########################################

sub find_OPTIMAL_KING_simple {
   #description _ finds a metric combination of (sub)optimal KING. (simple version)
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $maxK = 0;
   my $metricMaxK;
   my %Kmetrics;

   if ($verbose) {
      print STDERR "[IQMT] finding sub-optimal metric set...\n";
      print STDERR "(1) compute KING measure for each individual metric...\n";
   }
   #(1) compute KING measure for each individual metric
   foreach my $metric (@{$metrics}) {
      if ($verbose) { print STDERR "$metric"; }
      my @lmetric = ($metric);
      my $K = KING($IQ, \@lmetric, $topics, $peers, $models, $verbose);
      QARLA::unload($IQ, \@lmetric, $verbose);
      if ($K > $maxK){ $maxK = $K; $metricMaxK = $metric; }
      $Kmetrics{$metric} = $K;
      #$Kmetrics{$K}.=" $metric";
      if ($verbose) { print STDERR " [K = ", sprintf("%6.4f", $K), "]\n"; }
   }

   if ($verbose) {
      print STDERR "maxK = ", sprintf("%6.4f", $maxK), " ($metricMaxK)\n";
      print STDERR "(2) sorting metrics by KING...";
   }

   #(2) sort metrics by KING
   my @Kmetrics_sorted = sort { $Kmetrics{$b} <=> $Kmetrics{$a} } keys %Kmetrics;

   if ($verbose) {
      print STDERR " [DONE]\n";
      print STDERR "SORTED METRICS = ", join(" ", @Kmetrics_sorted), "\n";
   }

   my @SET;
   push(@SET, shift(@Kmetrics_sorted));
  
   #(3) process metrics in KING order
   if ($verbose) {
      print STDERR "(3) processing individual metrics...\n";
   }
   foreach my $metric (@Kmetrics_sorted) {
      if ($verbose) { print STDERR "[$metric] "; }
      push(@SET, $metric);
      if ($verbose) { print STDERR "set = ", join(" ", @SET), "\n"; }      
      my $K = KING($IQ, \@SET, $topics, $peers, $models, $verbose);
      if ($K > $maxK) {
         $maxK = $K;
         if ($verbose) { print STDERR "  ---> (K = ", sprintf("%6.4f", $K), ") \t\t ***** [ADDED] *****\n"; }
      }
      else {
         pop(@SET);
         my @lmetric = ($metric);
         QARLA::unload($IQ, \@lmetric, $verbose);
         if ($verbose) { print STDERR "  ---> (K = ", sprintf("%6.4f", $K), ")\n"; }
      }
   }

   return (\@SET, $maxK);
}

sub find_OPTIMAL_KING_TT_simple {
   #description _ finds a metric combination of (sub)optimal KING. (simple version)
   #param1  _ IQ structure
   #param2  _ set of metrics
   #param3  _ set of segments (topics)
   #param4  _ set of systems
   #param5  _ set of references (models)
   #param6  _ verbosity (0/1)

   my $IQ = shift;
   my $metrics = shift;
   my $topics = shift;
   my $peers = shift;
   my $models = shift;
   my $verbose = shift;

   my $maxK = 0;
   my $metricMaxK;
   my %Kmetrics;

   if ($verbose) {
      print STDERR "[IQMT] finding sub-optimal metric set...\n";
      print STDERR "(1) compute KINGtt measure for each individual metric...\n";
   }
   #(1) compute KING measure for each individual metric
   foreach my $metric (@{$metrics}) {
      if ($verbose) { print STDERR "$metric"; }
      my @lmetric = ($metric);
      my $K = KING_TT($IQ, \@lmetric, $topics, $peers, $models, $verbose);
      QARLA::unload($IQ, \@lmetric, $verbose);
      if ($K > $maxK){ $maxK=$K; $metricMaxK = $metric; }
      $Kmetrics{$metric} = $K;
      #$Kmetrics{$K}.=" $metric";
      if ($verbose) { print STDERR " [K = ", sprintf("%6.4f", $K), "]\n"; }
   }

   if ($verbose) {
      print STDERR "maxK = ", sprintf("%6.4f", $maxK), " ($metricMaxK)\n";
      print STDERR "(2) sorting metrics by KINGtt...";
   }

   #(2) sort metrics by KING
   my @Kmetrics_sorted = sort { $Kmetrics{$b} <=> $Kmetrics{$a} } keys %Kmetrics;

   if ($verbose) {
      print STDERR " [DONE]\n";
      print STDERR "SORTED METRICS = ", join(" ", @Kmetrics_sorted), "\n";
   }

   my @SET;
   push(@SET, shift(@Kmetrics_sorted));
  
   #(3) process metrics in KING order
   if ($verbose) {
      print STDERR "(3) processing individual metrics...\n";
   }
   foreach my $metric (@Kmetrics_sorted) {
      if ($verbose) { print STDERR "[$metric] "; }
      push(@SET, $metric);
      if ($verbose) { print STDERR "set = ", join(" ", @SET), "\n"; }
      my $K = KING_TT($IQ, \@SET, $topics, $peers, $models, $verbose);
      if ($K > $maxK) {
         $maxK = $K;
         if ($verbose) { print STDERR "  ---> (K = ", sprintf("%6.4f", $K), ") \t\t ***** [ADDED] *****\n"; }
      }
      else {
         pop(@SET);
         my @lmetric = ($metric);
         QARLA::unload($IQ, \@lmetric, $verbose);
         if ($verbose) { print STDERR "  ---> (K = ", sprintf("%6.4f", $K), ")\n"; }
      }
   }

   return (\@SET, $maxK);
}

1;
